/***********************************************************************/
/*                                                                     */
/*  FILE        :AD_RNG.c                                              */
/*  DATE        :2018/Feb/12 by KiriN                                  */
/*  DESCRIPTION :Main Program                                          */
/*  CPU TYPE    :RL78                                                  */
/*                                                                     */
/*  NOTE:Sample of random number generation with ADC                   */
/*                                                                     */
/***********************************************************************/
#include "iodefine.h" 

#define NUMBER_OF_SEEDS   8

//Function prototype
void main(void);
unsigned char rng(void);
unsigned char rn_test(void);
void init_adc(void);

//Global variables
unsigned char level; //Quality of random number generation function
unsigned long id;    //24bit random number
const unsigned char bitmask[8]={0x01,0x02,0x04,0x08,0x10,0x20,0x40,0x80}; //Bit position fast conversion table

void main(void)
{
	init_adc();
	level=rn_test();	//Quality of random number generation function(i̊mF)
	id=(((unsigned long)rng())<<16)+(((unsigned long)rng())<<8)+((unsigned long)rng());	//24bit random number

	while(1); //End of the program
}

void init_adc(void){
    PM2 |= 0x03; // set port P20-23 input 
    ADPC = 0x04; // set port P20-23 ANI0-ANI2
    ADCEN = 1;   // Ssupply AD clock
    ADM0 = 0x00; // Disable AD conversion and clear ADM0 register
    ADMK = 1;    // Disable INTAD interrupt
    ADIF = 0;    // Clear INTAD interrupt flag
    ADM0 = 0x30; // AD CONVERSION CLOCK=4, AD TIME MODE NORMAL1, AD OPERMODE SELECT
    ADM1 = 0x00; // AD TRIGGER SOFTWARE, AD CONVMODE CONSELECT
    ADM2 = 0x00; // AD POSITIVE VDD, AD NEGATIVE VSS, ADAREA MODE1, AD RESOLUTION 10BIT          
    ADUL = 0xFF; // AD ADUL VALUE=MAX
    ADLL = 0x00; // AD ADLL VALUE=MIN
    ADS =  0x02; // AD INPUT=CHANNEL2
    ADCE = 1;    // Enable AD comparator
    ADCS = 1;    // Enable AD (no wait start)
}

/*
Function 
	Generating Random Number
Parameter
	N/A
RETURN
	8bit Random Number
*/
unsigned char rng(void) {
	unsigned char rn=0; //8bit Random Number
	unsigned char cnt=0; //Temporary for compressing 8 bits to 1 bit
	unsigned char i,j;
	for(i=0;i<8;i++){
		for (j=0;j<NUMBER_OF_SEEDS;j++){
			while(ADIF==0);
			ADIF=0;
			/*Check the least significant bit of ADCR*/
			if((ADCR & 0x0040)==0){
				cnt=~cnt; //check XOR for N times A/D bit
			}
		}
		/*Make Random Number bit*/
		if(cnt==0){
			rn |= bitmask[i]; //Set bit
		}

	}
	return rn;
}

/*
Function
	Quality confirmation of bit variations
Parameter
	N/A
RETURN
	8bit Random Number
*/
unsigned char rn_test(void){
	unsigned int i;
	unsigned int l; //run length counter
	unsigned int rn, b, n; //random nuber & bit
	unsigned long long addv,tempv; //temporary value
	unsigned int  shiftv; //divider
	unsigned char retv; //return value
	unsigned int result[2][32]; //run length counter
	
	//init result
	for (i=0;i<32;i++){
	   result[0][i]=result[1][i]=0;
	}
	l=0;
	rn=rng(); //get 8bit Random Number
	b=((rn & 0x01)==0) ? 0x00:0x01;
	
	//Loop from [1], b as [0]
	for (i=1;i<20000;i++){
		n=((rn & bitmask[i & 0x0007])==0x00) ? 0x00:0x01;

		if(b!=n) {
        		result[b][(l>31) ? 31:l]++; //mark result
			l=0; //clear length
			b=n; //invert bit
		}else{
			l++;
		}

		if((i&0x0007)==0x0000){
			rn=rng(); //get new 8bit Random Number
		}	
	}

	if(l!=0){ //mark last loop result
		result[b][(l>31) ? 31:l]++; //mark result
	}

	// TRNG length vs Chi-squared Test
	// 
	//	TRNG     24bitTRNG    The test
	//	length   at 64units   statistic
	//	         match        of A/D RNG
	//
	//	24bit    0.01%            20
	//	23bit    0.02%            60
	//	22bit    0.05%           181
	//	21bit    0.10%           548
	//	20bit    0.19%          1660
	//	19bit    0.38%          5034
	//	18bit    0.77%         15259
	//	17bit    1.53%         46256
	//	16bit    3.03%        140222
	//	15bit    5.97%        425074
	//	14bit   11.59%       1288582
	//	13bit   21.87%       3906250
	//	12bit   39.03%      11841536
	//	11bit   63.02%      35896824
	//	10bit   86.61%     108818820
	//	 9bit   98.36%     329876978
	//	 8bit   99.98%    1000000000

	// Make the test statistic
	addv =0;
	shiftv =5;
	for(i=0;i<6;i++){
		tempv=(unsigned long long)result[0][i];
		tempv=(shiftv==0)?tempv:tempv>>(shiftv--);
		tempv=tempv=(tempv==0)? 0:(tempv<78)? 78-tempv : tempv-78;
		tempv=(tempv<0x20000000)? tempv : 0x20000000;//overflow check
		tempv*=tempv;
		addv+=tempv;
	}
	shiftv =1;
	for(i=6;i<32;i++){
		tempv=(unsigned long long)result[0][i]<<(shiftv++);
		tempv=(tempv==0)? 0:(tempv<78)? 78-tempv : tempv-78;
		tempv=(tempv<0x20000000)? tempv : 0x20000000;//overflow check
		tempv*=tempv;
		addv+=tempv;
	}
	
	shiftv =5;
	for(i=0;i<6;i++){
		tempv=(unsigned long long)result[1][i];
		tempv=(shiftv==0)?tempv:tempv>>(shiftv--);
		tempv=(tempv==0)? 0:(tempv<78)? 78-tempv : tempv-78;
		tempv=(tempv<0x20000000)? tempv : 0x20000000;//overflow check
		tempv*=tempv;
		addv+=tempv;
	}
	shiftv =1;
	for(i=6;i<32;i++){
		tempv=(unsigned long long)result[1][i]<<(shiftv++);
		tempv=(tempv==0)? 0:(tempv<78)? 78-tempv : tempv-78;
		tempv=(tempv<0x20000000)? tempv : 0x20000000;//overflow check
		tempv*=tempv;
		addv+=tempv;
	}
	//Compare with Squaring(non root) value
	if       (addv<=               400 ) { retv=24;
	} else if(addv<=              3600 ) { retv=23;
	} else if(addv<=             32761 ) { retv=22;
	} else if(addv<=            300304 ) { retv=21;
	} else if(addv<=           2755600 ) { retv=20;
	} else if(addv<=          25341156 ) { retv=19;
	} else if(addv<=         232837081 ) { retv=18;
	} else if(addv<=        2139617536 ) { retv=17;
	} else if(addv<=       19662209284 ) { retv=16;
	} else if(addv<=      180687905476 ) { retv=15;
	} else if(addv<=     1660443570724 ) { retv=14;
	} else if(addv<=    15258789062500 ) { retv=13;
	} else if(addv<=   140221974839296 ) { retv=12;
	} else if(addv<=  1288581973286980 ) { retv=11;
	} else if(addv<= 11841535586192400 ) { retv=10;
	} else if(addv<=108818820614412000 ) { retv= 9;    //56bit for long long
	} else                               { retv= 8;	

	}
	return retv;
}
